@@ -1,10 +1,13 @@
 package ShardedKV::Storage::MySQL;
 {
-  $ShardedKV::Storage::MySQL::VERSION = '0.08';
+  $ShardedKV::Storage::MySQL::VERSION = '0.09';
 }
 use Moose;
 # ABSTRACT: MySQL storage backend for ShardedKV
 
+use Time::HiRes qw(sleep);
+use Carp ();
+
 with 'ShardedKV::Storage';
 
 has 'mysql_master_connector' => (
@@ -25,9 +28,16 @@ has 'mysql_connection' => (
   builder => '_make_master_conn',
 );
 
+
 sub _make_master_conn {
   my $self = shift;
-  return $self->mysql_master_connector->();
+  my $dbh = $self->mysql_master_connector->();
+  if ($dbh) {
+    $dbh->{RaiseError} = 1;
+    $dbh->{PrintError} = 0;
+    #$dbh->{AutoCommit} = 1;
+  }
+  return $dbh;
 }
 
 has 'table_name' => (
@@ -93,6 +103,19 @@ has '_number_of_params' => (
   builder => '_calc_no_params',
 );
 
+has 'max_num_reconnect_attempts' => (
+  is => 'rw',
+  isa => 'Int',
+  default => 5,
+);
+
+# Note that exponential back-off is on by default
+has 'reconnect_interval' => (
+  is => 'rw',
+  isa => 'Num',
+  default => 1,
+);
+
 sub _calc_no_params {
   my $self = shift;
   return 1 + scalar(@{$self->value_col_names});
@@ -177,28 +200,53 @@ sub get_master_dbh {
   return $master_dbh;
 }
 
+sub _run_sql {
+  my ($self, $method, $query, @args) = @_;
+
+  my $iconn;
+  my $rv;
+  while (1) {
+    my $dbh = $self->get_master_dbh;
+    eval {
+      $rv = $dbh->$method($query, @args);
+      1
+    } or do {
+      my $err = $@ || 'Zombie error';
+      ++$iconn;
+      if ($err =~ /MySQL server has gone away/i
+          and $iconn <= $self->max_num_reconnect_attempts)
+      {
+        sleep($self->reconnect_interval * 2 ** ($iconn-2)) if $iconn > 1;
+        $self->refresh_connection;
+        redo;
+      }
+      Carp::confess("Despite trying hard: $err");
+    };
+    last;
+  }
+
+  return $rv;
+}
+
 sub get {
   my ($self, $key) = @_;
-
-  my $rv = $self->get_master_dbh->selectall_arrayref($self->get_query, undef, $key);
+  my $rv = $self->_run_sql('selectall_arrayref', $self->get_query, undef, $key);
   return ref($rv) ? $rv->[0] : undef;
 }
 
 sub set {
   my ($self, $key, $value_ref) = @_;
 
-  my $set_query = $self->set_query;
   Carp::croak("Need exactly " . ($self->{_number_of_params}-1) . " values, got " . scalar(@$value_ref))
     if not scalar(@$value_ref) == $self->{_number_of_params}-1;
 
-  my $rv = $self->get_master_dbh->do($set_query, undef, $key, @$value_ref);
+  my $rv = $self->_run_sql('do', $self->set_query, undef, $key, @$value_ref);
   return $rv ? 1 : 0;
 }
 
 sub delete {
   my ($self, $key) = @_;
-
-  my $rv = $self->get_master_dbh->do($self->delete_query, undef, $key);
+  my $rv = $self->_run_sql('do', $self->delete_query, undef, $key);
   return $rv ? 1 : 0;
 }
 
@@ -215,7 +263,7 @@ ShardedKV::Storage::MySQL - MySQL storage backend for ShardedKV
 
 =head1 VERSION
 
-version 0.08
+version 0.09
 
 =head1 SYNOPSIS
 
